from abc import ABC, abstractmethod

class Call:
    def __init__(self, id: str = "", name: str = "", arguments: dict = {}):
        self.id = id
        self.name = name
        self.arguments = arguments


class LLMResponse(ABC):
    @abstractmethod
    def get(self) -> str:
        return ""

    @abstractmethod
    def get_call(self) -> Call:
        return Call()

    @abstractmethod
    def is_text(self)-> bool:
        return True

    @abstractmethod
    def is_call(self) -> bool:
        return False



import json
from phoenix.domain import model, response, metadata
from phoenix.models.openai_history import ChatHistory
from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.models import ChatCompletions, SystemMessage
from azure.core.credentials import AzureKeyCredential

class AzureAIResponse(response.LLMResponse):
    def __init__(self, _message: ChatCompletions):
        self.message = _message

    def _get_message(self):
        return self.message.choices[-1].message

    def get(self) -> str:
        return self._get_message().content

    def is_text(self) -> bool:
        msg = self._get_message()
        return msg.content is not None and len(msg.content) > 0

    def get_call(self) -> response.Call:
        # Only one tool call is supported
        message = self.get_call_message()
        id = message['tool_calls'][0]['id']
        function = message['tool_calls'][0]['function']
        name = function['name']
        arguments = json.loads(function['arguments'])
        return response.Call(id, name, arguments)

    def get_call_message(self):
        return self.message.choices[-1]['message']

    def is_call(self) -> bool:
        message = self.get_call_message()
        if "tools_connector" in message and len(message['tool_calls']) > 0:
            return True
        return False


class AzureAIInferece(model.ConversationalAgent):
    system: str = ""
    tools: list = []

    def __init__(self, system: str = "", tools: list = [], history = ChatHistory(), *args, **kwargs):
        token = kwargs.get("token", "")
        self.system = system
        self.tools = [{
            "type": "function",
            "function": {
                "name": tool.name,
                "description": tool.description,
                "input_schema": tool.inputSchema
            }
        } for tool in tools]
        self.client = ChatCompletionsClient(
            endpoint="https://models.github.ai/inference",
            credential=AzureKeyCredential(token),
        )


        self.model = kwargs.get("model", "gpt-4o")
        self.max_tokens = kwargs.get("max_tokens", 4096)
        self.chat_history = history

    def with_system(self, system: str):
        self.system = system

    def with_tools(self, tools: list):
        self.tools = [{
            "type": "function",
            "function": {
                "name": tool.name,
                "description": tool.description,
                "input_schema": tool.inputSchema
            }
        } for tool in tools]

    def prompt(self, _message: str, _metadata: metadata.Metadata) -> AzureAIResponse:
        if len(_message) != 0:
            self.chat_history.push("user", _message, _metadata)

        messages: list = [SystemMessage(self.system)]
        for message in self.chat_history.get(_metadata):
            messages.append(message)

        tools = self.tools
        if len(tools) == 0:
            tools = None

        response = self.client.complete(
            messages=messages,
            model=self.model,
            temperature=1,
            max_tokens=self.max_tokens,
            top_p=1,
            tools=tools,
        )

        response = AzureAIResponse(response)

        if response.is_text():
            content = response.get()
            self.chat_history.push("assistant", content, _metadata)

        if response.is_call():
            call = response.get_call_message()
            self.chat_history.push("tool_call", call, _metadata)

        return response



from typing_extensions import Any
from phoenix.domain import tools_connector
from contextlib import AsyncExitStack
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

class MCPServer:
    # TODO: Support multiple languages
    def __init__(self, path: str):
        self.path = path
        self.command = "python3"


class MCPClient(tools_connector.ToolsConnector):
    def __init__(self, servers: list[MCPServer]):
        self.session: ClientSession
        self.exit_stack = AsyncExitStack()
        self.servers = servers
        self.tools = []

    async def connect(self):
        # TODO: Connect to multiple servers
        server = self.servers[0]

        server_params = StdioServerParameters(
            command=server.command,
            args=[server.path],
            env=None
        )

        stdio_transport = await self.exit_stack.enter_async_context(stdio_client(server_params))
        self.stdio, self.write = stdio_transport
        self.session = await self.exit_stack.enter_async_context(
            ClientSession(
                self.stdio, self.write
            )
        )
        await self.session.initialize()

        self.tools = await self._list()

    def list(self) -> list:
        return self.tools

    async def _list(self):
        response = await self.session.list_tools()
        return response.tools

    async def call(self, tool: str, args: dict) -> Any:
        result = await self.session.call_tool(tool, args)
        response = result.content[0]
        if response.type == "text":
            return response.text

        return None

    async def cleanup(self):
        await self.exit_stack.aclose()



from phoenix.domain import tools_connector

class DummyConnector(tools_connector.ToolsConnector):
    def list(self) -> list:
        return []

    async def call(self, tool: str, args: dict):
        return None



from phoenix.domain.agent import Agent as CoreAgent
from phoenix.domain.metadata import Metadata
from phoenix.domain.history import History
from phoenix.domain.model import ConversationalAgent
from phoenix.domain.tools_connector import ToolsConnector
from phoenix.connectors.dummy import DummyConnector

class Agent(CoreAgent):
    def __init__(self,
        brain: ConversationalAgent,
        history: History,
        connector: ToolsConnector = DummyConnector(),
        system: str = "",
        tools: list = []
    ):
        self.brain = brain
        self.tools = tools
        self.history = history
        self.system = system
        self.connector = connector

        self.brain.with_system(self.system)
        self.brain.with_tools(self.tools)

        if self.connector:
            tools = self.connector.list()
            self.brain.with_tools(tools)

    async def call(self, _query: str, _metadata: Metadata) -> str:
        out = ""

        while True:
            response = self.brain.prompt(_query, _metadata)

            if response.is_text():
                out += response.get()
            if response.is_call():
                call = response.get_call()
                result = await self.connector.call(call.name, call.arguments)
                self.history.push("tool", {
                    "name": call.name,
                    "id": call.id,
                    "content": result
                }, _metadata)
                response = self.brain.prompt("", _metadata)
                out += "\"Figuring things out\""

        return out



from abc import ABC, abstractmethod
from phoenix.domain.response import LLMResponse
from phoenix.domain.metadata import Metadata

class ConversationalAgent(ABC):
    @abstractmethod
    def with_system(self, system: str):
        pass

    @abstractmethod
    def with_tools(self, tools: list):
        pass

    @abstractmethod
    def prompt(self, _message: str, _metadata: Metadata) -> LLMResponse:
        pass



from phoenix.domain import history, metadata

class ChatHistory(history.History):
    def __init__(self):
        self.histories = {}

    def push(self, message_type, message, _metadata: metadata.Metadata):
        if _metadata.get_id() not in self.histories:
            self.histories[_metadata.get_id()] = []

        if message_type == "user" or message_type == "assistant":
            self.histories[_metadata.get_id()].append({
                    "role": message_type,
                    "content": message
            })
        elif message_type == "tool":
            self.histories[_metadata.get_id()].append({
                    "role": "tool",
                    "tool_call_id": message["id"],
                    "content": message["content"],
                    "name": message["name"]
            })
        else:
            self.histories[_metadata.get_id()].append(message)

    def get(self, _metadata: metadata.Metadata) -> list:
        return self.histories[_metadata.get_id()]



import phoenix.domain.metadata as metadata
import uuid

class UserSession(metadata.Metadata):
    def __init__(self):
        self.id = str(uuid.uuid4())

    def get_id(self) -> str:
        return self.id

    def to_string(self) -> str:
        return ""



from phoenix.domain import model, response
from anthropic import Anthropic
from anthropic.types.message import Message  as ResponseMessage
from anthropic.types.text_block import TextBlock
from anthropic.types.message_param import MessageParam

class AnthropicResponse(response.LLMResponse):
    def __init__(self, _message: ResponseMessage):
        self.message = _message

    def get(self) -> str:
        if self.is_text() and isinstance(self.message.content[-1], TextBlock):
            msg = self.message.content[-1]
            return msg.text

        return ""

    def is_text(self) -> bool:
        return isinstance(self.message.content[-1], TextBlock)

    def is_call(self) -> bool:
        last = self.message.content[-1]
        return last is not TextBlock

    def get_call(self) -> tuple[str, str]:
        return "", ""

class AnthropicLLM(model.ConversationalAgent):
    system: str = ""
    tools: list = []

    def __init__(self, system: str = "", tools: list = [], *args, **kwargs):
        self.system = system
        self.tools = tools
        self.model = kwargs.get("model", "claude-3-5-sonnet-20241022")
        self.max_tokens = kwargs.get("max_tokens", 1000)
        self._anthropic = Anthropic(
            api_key = kwargs.get("api_key", ""),
            base_url = kwargs.get("base_url", "")
        )

    def with_system(self, system: str):
        self.system = system

    def with_tools(self, tools: list):
        self.tools = tools


    def _get_messages(self, messages: list[str]) -> list[MessageParam]:
        return [{"role": "user", "content": msg} for msg in messages]

    def prompt(self, *_messages: str) -> AnthropicResponse:
        messages = self._get_messages(list(_messages))

        response = self._anthropic.messages.create(
            model=self.model,
            max_tokens=self.max_tokens,
            messages=messages,
            tools=self.tools,
            system=self.system,
        )

        return AnthropicResponse(response)



from abc import ABC, abstractmethod
from phoenix.domain.metadata import Metadata

class Agent(ABC):
    @abstractmethod
    async def call(self, _query: str, _metadata: Metadata) -> str:
        pass



from abc import ABC, abstractmethod
from phoenix.domain.metadata import Metadata

class History(ABC):
    @abstractmethod
    def push(self, message_type, message, _metadata: Metadata):
        pass

    @abstractmethod
    def get(self, _metadata: Metadata) -> list:
        pass



from abc import ABC, abstractmethod
from typing_extensions import Any

class ToolsConnector(ABC):
    @abstractmethod
    def list(self) -> list:
        pass

    @abstractmethod
    async def call(self, tool: str, args: dict) -> Any:
        pass
